import pandas as pd

import pemi

class SaSqlSourcePipe(pemi.Pipe):
    def __init__(self, *, sql, engine, schema=None, result=True, chunk_size=None):
        super().__init__()

        self.sql = sql
        self.engine = engine
        self.schema = schema
        self.result = result
        self.chunk_size = chunk_size

        self.target(
            pemi.PdDataSubject,
            name='main',
            schema=self.schema
        )


    def _get_result(self, conn):
        sql_df = pd.DataFrame()
        if self.chunk_size is None:
            sql_df = pd.read_sql(self.sql, conn)
        else:
            for chunk in pd.read_sql(self.sql, conn, chunksize=self.chunk_size):
                sql_df = sql_df.append(chunk, ignore_index=True)
        return sql_df


    def extract(self):
        pemi.log.info("Executing SQL '%s' via:\n%s", self.name, self.sql)

        data = None
        with self.engine.connect() as conn:
            if self.result:
                data = self._get_result(conn)
            else:
                conn.execute(self.sql)

        return data

    def parse(self, data):
        pemi.log.info("Parsing '%s' results", self.name)

        if data is None:
            return None

        if self.schema is None:
            self.targets['main'].df = data
        else:
            mapper = data.mapping(
                [(name, name, field.coerce) for name, field in self.schema.items()],
                on_error='raise'
            )
            self.targets['main'].df = mapper.mapped

        return self.targets['main'].df

    def flow(self):
        self.parse(self.extract())



from IPython.display import display
from graphviz import Digraph


GRAPH_OPTS = {
    'source': {'shape': 'invhouse', 'color': 'lightgreen', 'style': 'filled'},
    'target': {'shape': 'house', 'color': 'lightblue', 'style': 'filled'},
    'pipe': {'shape': 'box', 'color': 'orange', 'style': 'filled, rounded'}
}

def graph(pipe):
    '''
    Running this in a Jupyter notebook will draw a graph of the connections between
    sources, targets, and pipes.
    '''


    dot = Digraph()

    for pname in pipe.pipes:
        if pname != 'self':
            dot.node(pname, **GRAPH_OPTS['pipe'])

        for source in pipe.pipes[pname].sources:
            if pname == 'self':
                label = 'self[{}]'.format(source)
            else:
                label = source

                dot.edge('s.{}[{}]'.format(pname, source), pname)


            dot.node(
                's.{}[{}]'.format(pname, source),
                label=label,
                **GRAPH_OPTS['source']
            )



        for target in pipe.pipes[pname].targets:
            if pname == 'self':
                label = 'self[{}]'.format(target)
            else:
                label = target

                dot.edge(pname, 't.{}[{}]'.format(pname, target))

            dot.node(
                't.{}[{}]'.format(pname, target),
                label=label,
                **GRAPH_OPTS['target']
            )


    for conn in pipe.connections.connections:
        from_ind = 't'
        to_ind = 's'
        if conn.from_pipe_name == 'self':
            from_ind = 's'
        if conn.to_pipe_name == 'self':
            to_ind = 't'

        dot.edge(
            '{}.{}[{}]'.format(from_ind, conn.from_pipe_name, conn.from_subject_name),
            '{}.{}[{}]'.format(to_ind, conn.to_pipe_name, conn.to_subject_name),
            arrowhead='dot', arrowtail='dot', dir='both'
        )

    display(dot)



import pickle
import copy
from collections import OrderedDict

import pemi
import pemi.connections


class Pipe:
    '''
    A pipe is a parameterized collection of sources and targets which can be executed (flow).


    Args:
        name (str): Assign a name to the pipe
        **params: Additional keyword parameters

    Attributes:
        name (str): The name of the pipe.
        sources (dict): A dictionary where the keys are the names of source data subjects\
                        and the values are instances of a data subject class.
        targets (dict): A dictionary where the keys are the names of target data subjects\
                        and the values are instances of a data subject class.
        pipes (dict): A dictionary referencing nested pipes where the keys are the names\
                      of the nested pipes and the values are the nested pipe instances.
                      All pipes come with at least one nested pipe called 'self'.
        connections (PipeConnection): Pemi connection object.

    '''

    def __init__(self, *, name='self', **params):

        self.name = name
        self.params = params
        self.sources = OrderedDict()
        self.targets = OrderedDict()
        self.pipes = OrderedDict()
        self.connections = pemi.connections.PipeConnections()

        self.pipes['self'] = self

    def source(self, subject_class, name, schema=None, **kwargs):
        '''
        Define a source data subject for this pipe.

        Args:
            subject_class (class): The :class:`DataSubject` class this source uses.
            name (str) : Name of this data subject.
            schema (schema): :class:`Schema` associated with this source.

        Example:
            Creating a source::

                class MyPipe(pemi.Pipe):
                    def __init__(self):
                        super().__init__()

                        self.source(
                            pemi.PdDataSubject,
                            name='main'
                        )

            >>> MyPipe().sources.keys()
            ['main']
        '''

        schema = schema or pemi.Schema()
        self.sources[name] = subject_class(
            pipe=self,
            name=name,
            schema=schema,
            **kwargs
        )

    def target(self, subject_class, name, schema=None, **kwargs):
        '''
        Define a target data subject for this pipe.

        Args:
            subject_class (class): The :class:`DataSubject` class this target uses.
            name (str) : Name of this data subject.
            schema (schema): :class:`Schema` associated with this target.

        Example:
            Creating a target::

                class MyPipe(pemi.Pipe):
                    def __init__(self):
                        super().__init__()

                        self.target(
                            pemi.PdDataSubject,
                            name='main'
                        )

            >>> MyPipe().targets.keys()
            ['main']
        '''

        schema = schema or pemi.Schema()
        self.targets[name] = subject_class(
            pipe=self,
            name=name,
            schema=schema,
            **kwargs
        )

    def pipe(self, name, pipe):
        '''
        Defines a named pipe nested in this pipe.

        Args:
            name (str): Name of the nested pipe.
            pipe (pipe): The nested pipe instance.

        Example:
            Creating a target::

                class MyPipe(pemi.Pipe):
                    def __init__(self):
                        super().__init__()

                        self.pipe(
                            name='nested',
                            pipe=pemi.Pipe()
                        )

            >>> MyPipe().pipes.keys()
            ['self', 'nested']
        '''

        pipe.name = name
        self.pipes[name] = pipe


    def connect(self, from_pipe_name, from_subject_name):
        '''
        Connect one nested pipe to another

        Args:
            from_pipe_name (str): Name of the nested pipe that contains the\
                source of the connection.
            from_subject_name (str): Name of the data subject in the nested pipe that\
                contains the source of the connection.  This data subject needs to be a *target*
                of the pipe referenced by `from_pipe_name`.

        Returns:
            :class:`PipeConnection`: the PipeConnection object.

        Example:
            Connecting the target of one pipe to the source of another::

                class MyPipe(pemi.Pipe):
                    def __init__(self):
                        super().__init__()

                        self.pipe(
                            name='get_awesome_data',
                            pipe=GetAwesomeDataPipe()
                        )
                        self.connect('get_awesome_data', 'main').to('load_awesome_data', 'main')

                        self.pipe(
                            name='load_awesome_data',
                            pipe=LoadAwesomeDataPipe()
                        )

        '''

        conn = pemi.connections.PipeConnection(self, from_pipe_name, from_subject_name)

        self.connections.append(conn)
        return conn


    def to_pickle(self, picklepipe=None):
        '''
        Recursively pickle all of the data subjects in this and all nested pipes

        Args:
            picklepipe: A pickled representation of a pipe.  Only used for recursion\
                           not meant to be set by user.

        Returns:
            bytes: A bytes object containing the pickled pipe.

        Example:
            Pickling a pipe:

            >>> MyPipe.to_pickle()
            b'.. <bytes> ..'


        '''
        picklepipe = picklepipe or Pipe()

        for name, source in self.sources.items():
            psource = copy.copy(source)
            psource.pipe = picklepipe
            picklepipe.sources[name] = psource

        for name, target in self.targets.items():
            ptarget = copy.copy(target)
            ptarget.pipe = picklepipe
            picklepipe.targets[name] = ptarget

        for name, nestedpipe in self.pipes.items():
            if nestedpipe == self: continue

            nestedpicklepipe = Pipe()
            picklepipe.pipe(name=name, pipe=nestedpicklepipe)
            picklepipe.pipes[name] = nestedpipe.to_pickle(nestedpicklepipe)

        return pickle.dumps(picklepipe)



    def from_pickle(self, picklepipe=None):
        '''
        Recursively load all data subjects in all nested pipes from a pickled bytes object
        created by `to_pickle`.

        Args:
            picklepipe: The bytes object created by `to_pickle`

        Returns:
            self:

        Example:
            De-pickling a pickled pipe::

                my_pipe = MyPipe()
                pickled = my_pipe.to_pickle()

                my_other_pipe = MyPipe().from_pickle(pickled)
        '''

        picklepipe = pickle.loads(picklepipe)

        for name, source in picklepipe.sources.items():
            source.pipe = self
            self.sources[name] = source

        for name, target in picklepipe.targets.items():
            target.pipe = self
            self.targets[name] = target

        for name, nestedpicklepipe in picklepipe.pipes.items():
            if nestedpicklepipe == picklepipe: continue

            self.pipes[name].from_pickle(nestedpicklepipe)

        return self


    def flow(self):
        '''
        Execute this pipe.  This method is meant to be defined in a child class.

        Example:
            A simple hello-world pipe::

                class MyPipe(pemi.Pipe):
                    def flow(self):
                        print('hello world')

            >>> MyPipe().flow()
            'hello world'
        '''

        raise NotImplementedError

    def __str__(self):
        return "<{}({}) {}>".format(self.__class__.__name__, self.name, id(self))


class MockPipe(Pipe):
    def flow(self):
        pemi.log.debug('FLOWING mocked pipe: %s', self)

def mock_pipe(parent_pipe, pipe_name):
    pipe = parent_pipe.pipes[pipe_name]
    mocked = MockPipe(name=pipe.name)
    for source in pipe.sources:
        mocked.sources[source] = pipe.sources[source]
        mocked.sources[source].pipe = mocked

    for target in pipe.targets:
        mocked.targets[target] = pipe.targets[target]
        mocked.targets[target].pipe = mocked

    parent_pipe.pipes[pipe_name] = mocked



from collections import defaultdict

import dask
import dask.dot

import pemi

class DagValidationError(Exception): pass

class PipeConnection:
    def __init__(self, parent, from_pipe_name, from_subject_name):
        self.parent = parent
        self.from_pipe_name = from_pipe_name
        self.from_subject_name = from_subject_name
        self.group = None
        self.to_pipe_name = None
        self.to_subject_name = None


    def to(self, to_pipe_name, to_subject_name): #pylint: disable=invalid-name
        self.to_pipe_name = to_pipe_name
        self.to_subject_name = to_subject_name
        return self

    @property
    def from_pipe(self):
        return self.parent.pipes[self.from_pipe_name]

    @property
    def to_pipe(self):
        return self.parent.pipes[self.to_pipe_name]

    @property
    def from_subject(self):
        if self.from_pipe_name == 'self':
            return self.from_pipe.sources[self.from_subject_name]

        return self.from_pipe.targets[self.from_subject_name]

    @property
    def to_subject(self):
        if self.to_pipe_name == 'self':
            return self.to_pipe.targets[self.to_subject_name]

        return self.to_pipe.sources[self.to_subject_name]

    @property
    def is_from_self(self):
        return self.parent is self.from_pipe

    @property
    def is_to_self(self):
        return self.parent is self.to_pipe

    def connect(self):
        self.to_subject.connect_from(self.from_subject)

    def group_as(self, name):
        self.group = name
        return self

    def __str__(self):
        return 'PipeConnection: {}.{} -> {}.{}'.format(
            self.from_pipe,
            self.from_subject,
            self.to_pipe,
            self.to_subject
        )

    def __repr__(self):
        return '<{}>'.format(self.__str__())

class DaskPipe: #pylint: disable=too-few-public-methods
    'DaskPipe is just a wrapper around Pipe that allows us to use the Pipes in the context of Dask'

    def __init__(self, pipe, flow=True):
        self.pipe = pipe
        self.flow = flow

    def __call__(self, *args):
        pemi.log.info('DaskPipe flowing pipe %s', self.pipe)
        if self.flow:
            self.pipe.flow()
        return self

    def __str__(self):
        return self.pipe.name


class PipeConnections:
    def __init__(self, connections=None):
        self.connections = connections or []

    def append(self, conn):
        self.connections.append(conn)

    def group(self, group):
        return self.__class__([c for c in self.connections if c.group == group])

    def flow(self, dask_get=dask.get):
        dask_dag = self.dask_dag()
        return dask_get(dask_dag, list(dask_dag.keys()))


    def graph(self):
        return self.dask_graph()

    def dask_graph(self):
        return dask.dot.dot_graph(self.dask_dag(), rankdir='TB')

    def dask_dag(self):
        self.validate_dag()
        dag = {}
        for conn in self.connections:
            for node, edge in self._node_edge(conn).items():
                if node in dag:
                    dag[node][1].extend(edge[1])
                else:
                    dag[node] = edge

        return dag


    def validate_dag(self):
        targets = defaultdict(list)
        sources = defaultdict(list)
        for conn in self.connections:
            to_str = '{}.sources[{}]'.format(conn.to_pipe_name, conn.to_subject_name)
            from_str = '{}.targets[{}]'.format(conn.from_pipe_name, conn.from_subject_name)
            sources[from_str].append(to_str)
            targets[to_str].append(from_str)

            from_type = 'target'
            from_subjects = conn.parent.pipes[conn.from_pipe_name].targets
            if conn.is_from_self:
                from_type = 'source'
                from_subjects = conn.parent.pipes[conn.from_pipe_name].sources

            to_type = 'source'
            to_subjects = conn.parent.pipes[conn.to_pipe_name].sources
            if conn.is_to_self:
                to_type = 'target'
                to_subjects = conn.parent.pipes[conn.to_pipe_name].targets


            if not conn.from_pipe_name in conn.parent.pipes:
                msg = "Pipe '{}' not defined in parent pipe '{}'".format(
                    conn.from_pipe_name, conn.parent.name
                )
                raise DagValidationError(msg)
            if not conn.to_pipe_name in conn.parent.pipes:
                msg = "Pipe '{}' not defined in parent pipe '{}'".format(
                    conn.to_pipe_name, conn.parent.name
                )
                raise DagValidationError(msg)
            if not conn.from_subject_name in from_subjects:
                msg = "Pipe '{}' has no {} named '{}'".format(
                    conn.from_pipe_name, from_type, conn.from_subject_name
                )
                raise DagValidationError(msg)
            if not conn.to_subject_name in to_subjects:
                msg = "Pipe '{}' has no {} named '{}'".format(
                    conn.to_pipe_name, to_type, conn.to_subject_name
                )
                raise DagValidationError(msg)

        target_dupes = {k:v for k, v in targets.items() if len(v) > 1}
        if len(target_dupes) > 0:
            msg = 'Multiple connections to the same target.  ' \
                + 'Use a concatenator pipe instead.  Details: {}'.format(target_dupes)
            raise DagValidationError(msg)

        source_dupes = {k:v for k, v in sources.items() if len(v) > 1}
        if len(source_dupes) > 0:
            msg = 'Multiple connections from the same source.  ' \
                + 'Use a fork pipe instead.  Details: {}'.format(source_dupes)
            raise DagValidationError(msg)

    @staticmethod
    def _connect_to(source, connection):
        def __connect_to(target):
            pemi.log.debug('connecting %s to %s', target, source)
            connection.connect()
            return target
        __connect_to.__name__ = 'connect_to'
        return __connect_to

    @staticmethod
    def _get_subject(dstype, name):
        def __get_subject(daskpipe):
            pemi.log.debug('Getting %s[%s] from pipe %s', dstype, name, daskpipe.pipe)
            return getattr(daskpipe.pipe, dstype)[name]
        __get_subject.__name__ = '[{}]'.format(name)
        return __get_subject

    def _node_edge(self, conn):
        '''
        A dask graph is a dictionary mapping keys to computations.

        Here, the keys are either a collection of data subjects (sources or targets),
        or a specific subject.  Computations are either flowing a pipe, connecting
        a data subject of one pipe to a data subject of another pipe, or getting a specific
        data subject from a collection.

        This method builds a graph needed by dask to execute according to the right topology.
        '''

        keys = {}
        tasks = {}

        from_type = 'targets'
        to_type = 'sources'

        if conn.is_from_self:
            from_type = 'sources'
        if conn.is_to_self:
            to_type = 'targets'


        keys = {
            'from_pipe_subjects': '{pipe}.{dstype}'.format(
                pipe=conn.from_pipe_name,
                dstype=from_type
            ),
            'from_pipe_subject': '{pipe}.{dstype}[{subject}]'.format(
                pipe=conn.from_pipe_name,
                dstype=from_type,
                subject=conn.from_subject_name
            ),
            'to_pipe_subject': '{pipe}.{dstype}[{subject}]'.format(
                pipe=conn.to_pipe_name,
                dstype=to_type,
                subject=conn.to_subject_name
            ),
            'to_pipe_results': '{pipe}.targets'.format(
                pipe=conn.to_pipe_name
            ),
        }

        tasks = {
            'flow_from_pipe': (DaskPipe(conn.from_pipe, flow=(not conn.is_from_self)), []),
            'get_subject': (
                self._get_subject(from_type, conn.from_subject_name),
                keys['from_pipe_subjects']
            ),
            'connect_data': (
                self._connect_to(getattr(conn.to_pipe, to_type)[conn.to_subject_name], conn),
                keys['from_pipe_subject']
            ),
            'flow_to_pipe': (DaskPipe(conn.to_pipe), [keys['to_pipe_subject']])
        }


        if conn.is_to_self:
            del keys['to_pipe_results']
            del tasks['flow_to_pipe']

        graph_outline = {
            'from_pipe_subjects': 'flow_from_pipe',
            'from_pipe_subject': 'get_subject',
            'to_pipe_subject': 'connect_data',
            'to_pipe_results': 'flow_to_pipe'
        }

        return {
            keys[key]: tasks[task] for key, task in graph_outline.items() if key in keys
        }



import os
import sys
from pandas import pandas as pd
from tabulate import tabulate as tbl

_PATH = os.path.abspath(__file__)
_DIR_PATH = os.path.dirname(_PATH)
sys.path.append(_DIR_PATH)


class PemiTabular:
    __DEFAULT_FILE_DIR = 'pemi-tabular/'
    __DEFAULT_FILE_NAME = 'tabular.html'
    __CSS_FILE_PATH = _DIR_PATH + '/pemi-web/css/pemi_errors.css'

    def __init__(self, file_dir=__DEFAULT_FILE_DIR):
        super().__init__()
        self.__file_dir = file_dir
        self.__css_style = self.__load_css_file()
        self.__data = []
        self.__output_file = None

    @staticmethod
    def to_html_table(df, title):
        try:
            if not isinstance(df, pd.core.frame.DataFrame):
                raise ValueError
        except ValueError:
            raise ValueError("df argument should have a type of pandas.core.frame.DataFrame")

        html_title = "<h3 style=\"text-align: center;\">{}</h3> \n".format(title)
        df = df.applymap(str)
        df.replace("<", "&lt;", regex=True, inplace=True)
        df.replace(">", "&gt;", regex=True, inplace=True)
        html_table = html_title + tbl(df, list(df.columns.values), 'html', stralign='center')
        return html_table

    def add(self, df, df_name=None):
        if df_name is None:
            df_name = self.__generate_df_name()
        self.__data.append({
            "name": df_name,
            "df": df
        })

    def render(self, file=__DEFAULT_FILE_NAME):
        file_path = self.__file_dir + file
        self.__open(file_path)
        self.__append("<html><head>" + "<style>" + self.__css_style + "</style></head>")
        for _, item in enumerate(self.__data):
            html_table = PemiTabular.to_html_table(df=item['df'], title=item['name'])
            self.__append(html_table)
        self.__append("</html>")
        self.__close()

    def reset(self):
        self.__data = []

    def __append(self, data):
        self.__output_file.write(data)

    def __close(self):
        try:
            if self.__output_file is not None:
                self.__output_file.close()
        except IOError:
            raise IOError('Cannot close output file')

    def __open(self, file_path):
        self.__create_dir()

        try:
            self.__output_file = open(file_path, "w")
        except IOError:
            raise IOError('Cannot open file with path ' + file_path)

    def __generate_df_name(self):
        default_table_name = 'Pemi Errors Table {}'
        return default_table_name.format(len(self.__data) + 1)

    def __create_dir(self):
        try:
            if not os.path.exists(self.__file_dir):
                os.mkdir(self.__file_dir)
        except IOError:
            raise IOError('Cannot create directory ' + self.__file_dir)

    def __load_css_file(self):
        try:
            css_file = open(self.__CSS_FILE_PATH, "r")
            css_content = css_file.read()
        except IOError:
            raise IOError('Cannot find CSS file with path ' + self.__CSS_FILE_PATH)

        return css_content



import math
import numpy as np
import pandas as pd

def _isnan(value):
    try:
        return math.isnan(value)
    except TypeError:
        return None

def _isnat(value):
    try:
        return bool(np.isnat(value))
    except TypeError:
        return None

def isblank(value):
    isnan = _isnan(value)
    isnat = _isnat(value)

    if isnan is not None: return isnan
    if isnat is not None: return isnat

    blank = [np.nan, pd.NaT, [], {}, None, '']
    return value in blank


def concatenate(delimiter=''):
    def _concatenate(row):
        return delimiter.join(row)

    return _concatenate


def nvl(default=''):
    def _nvl(row):
        return next((v for v in row if not isblank(v)), default)

    return _nvl



import pemi

class SourcePipe(pemi.Pipe):
    '''
    A source pipe extracts data from some external source, and parses
    it into a data structure that can be used by subsequent pipes.
    '''
    def __init__(self, *, schema, **params):
        super().__init__(**params)

        self.schema = schema

        self.target(
            pemi.PdDataSubject,
            name='main',
            schema=self.schema
        )

        self.target(
            pemi.PdDataSubject,
            name='errors',
            schema=self.schema
        )

    def extract(self):
        # e.g., S3SourceExtractor.extract()
        # return extracted_data_that_can_be_parsed
        raise NotImplementedError

    def parse(self, data):
        # e.g., CsvParser.parse(data)
        # return parsed_data
        raise NotImplementedError

    def flow(self):
        self.parse(self.extract())


class TargetPipe(pemi.Pipe):
    '''
    A target pipe takes data provided to it, encodes it into a structure that can be
    understand by some external target, and then loads the data into that external target.
    '''

    def __init__(self, *, schema, **params):
        super().__init__(**params)

        self.schema = schema

        self.source(
            pemi.PdDataSubject,
            name='main',
            schema=self.schema
        )

        self.target(
            pemi.PdDataSubject,
            name='errors'
        )

        self.target(
            pemi.PdDataSubject,
            name='response'
        )

    def encode(self):
        # e.g., CsvTargetEncoder.encode()
        # return source_main_data_encoded_for_loader
        raise NotImplementedError

    def load(self, encoded_data):
        # e.g., S3Loader.load()
        # return results_from_load_operation
        raise NotImplementedError

    def flow(self):
        self.load(self.encode())


class ForkPipe(pemi.Pipe):
    ''' A fork pipe accepts a single source and delivers it to multiple named targets '''
    def __init__(self, *, subject_class=pemi.PdDataSubject, forks=None, **params):
        super().__init__(**params)
        forks = forks or []

        self.source(subject_class, name='main')
        for fork in forks:
            self.target(subject_class, name=fork)

    def flow(self):
        raise NotImplementedError


class ConcatPipe(pemi.Pipe):
    ''' A concat pipe accepts multiple sources and combines them into a single target '''
    def __init__(self, subject_class=pemi.PdDataSubject, sources=None, **params):
        super().__init__(**params)
        sources = sources or []
        self.named_sources = sources

        for source in sources:
            self.source(subject_class, name=source)
        self.target(subject_class, name='main')

    def flow(self):
        raise NotImplementedError



import json

import pandas as pd
import sqlalchemy as sa

import pemi
from pemi.fields import *

__all__ = [
    'PdDataSubject',
    'SaDataSubject',
    'SparkDataSubject'
]

class MissingFieldsError(Exception): pass

class DataSubject:
    '''
    A data subject is mostly just a schema and a generic data object
    Actually, it's mostly just a schema that knows which pipe it belongs to (if any)
    and can be converted from and to a pandas dataframe (really only needed for testing to work)
    '''

    def __init__(self, schema=None, name=None, pipe=None):
        self.schema = schema or pemi.Schema()
        self.name = name
        self.pipe = pipe

    def __str__(self):
        subject_str = '<{}({}) {}>'.format(self.__class__.__name__, self.name, id(self))
        if self.pipe:
            return '{}.{}'.format(self.pipe, subject_str)
        return subject_str

    def to_pd(self):
        raise NotImplementedError

    def from_pd(self, df, **kwargs):
        raise NotImplementedError

    def connect_from(self, _other):
        self.validate_schema()
        raise NotImplementedError

    def validate_schema(self): #pylint: disable=no-self-use
        return True


class PdDataSubject(DataSubject):
    def __init__(self, df=None, strict_match_schema=False, **kwargs):
        super().__init__(**kwargs)

        if df is None or df.shape == (0, 0):
            df = self._empty_df()
        self.strict_match_schema = strict_match_schema
        self.df = df

    def to_pd(self):
        return self.df

    def from_pd(self, df, **kwargs):
        self.df = df

    def connect_from(self, other):
        if other.df is None or other.df.shape == (0, 0):
            self.df = self._empty_df()
        else:
            self.df = other.df
        self.validate_schema()

    def validate_schema(self):
        'Verify that the dataframe contains all of the columns specified in the schema'
        if self.strict_match_schema:
            return self.validate_data_frame_columns()
        missing = set(self.schema.keys()) - set(self.df.columns)
        if len(missing) == 0:
            return True
        raise MissingFieldsError('DataFrame missing expected fields: {}'.format(missing))

    def validate_data_frame_columns(self):
        'Verify that the schema contains all the columns specefied in the dataframe'
        missing = set(self.df.columns) - set(self.schema.keys())
        if len(missing) > 0:
            raise MissingFieldsError("Schema is missing current columns: {}".format(missing))
        return True

    def _empty_df(self):
        return pd.DataFrame(columns=self.schema.keys())

class SaDataSubject(DataSubject):
    def __init__(self, engine, table, sql_schema=None, **kwargs):
        super().__init__(**kwargs)
        self.engine = engine
        self.table = table
        self.sql_schema = sql_schema

        self.cached_test_df = None

    def to_pd(self):
        if self.cached_test_df is not None:
            return self.cached_test_df

        with self.engine.connect() as conn:
            sql_df = pd.read_sql_table(
                self.table,
                conn,
                schema=self.sql_schema,
                columns=self.schema.keys()
            )

            df = pd.DataFrame()
            for column in list(sql_df):
                df[column] = sql_df[column].apply(self.schema[column].coerce)

            self.cached_test_df = df
        return df

    def from_pd(self, df, **to_sql_opts):
        self.cached_test_df = df
        pemi.log.debug('loading SaDataSubject with:\n%s', self.cached_test_df)

        to_sql_opts['if_exists'] = to_sql_opts.get('if_exists', 'append')
        to_sql_opts['index'] = to_sql_opts.get('index', False)
        if self.sql_schema:
            to_sql_opts['schema'] = self.sql_schema

        df_to_sql = df.copy()
        for field in self.schema.values():
            if isinstance(field, JsonField):
                df_to_sql[field.name] = df_to_sql[field.name].apply(json.dumps)

        with self.engine.connect() as conn:
            df_to_sql.to_sql(self.table, conn, **to_sql_opts)

    def connect_from(self, _other):
        self.engine.dispose()
        self.validate_schema()

    def __getstate__(self):
        return (
            [],
            {
                'url': self.engine.url,
                'table': self.table,
                'sql_schema': self.sql_schema
            }
        )

    def __setstate__(self, state):
        _args, kwargs = state
        self.engine = sa.create_engine(kwargs['url'])
        self.table = kwargs['table']
        self.sql_schema = kwargs['sql_schema']



class SparkDataSubject(DataSubject):
    def __init__(self, spark, df=None, **kwargs):
        super().__init__(**kwargs)
        self.spark = spark
        self.df = df

        self.cached_test_df = None

    def to_pd(self):
        if self.cached_test_df is not None:
            return self.cached_test_df

        converted_df = self.df.toPandas()
        self.cached_test_df = pd.DataFrame()
        for column in list(converted_df):
            self.cached_test_df[column] = converted_df[column].apply(self.schema[column].coerce)

        return self.cached_test_df

    def from_pd(self, df, **kwargs):
        self.df = self.spark.createDataFrame(df)

    def connect_from(self, other):
        self.spark = other.spark.builder.getOrCreate()
        self.df = other.df
        self.validate_schema()



import os

import pytest
import factory

import sqlalchemy as sa

import pemi
import pemi.testing as pt
import pemi.pipes.sa
from pemi.fields import *

sa_engine = sa.create_engine('postgresql://{user}:{password}@{host}/{dbname}'.format(
    user=os.environ.get('POSTGRES_USER'),
    password=os.environ.get('POSTGRES_PASSWORD'),
    host=os.environ.get('POSTGRES_HOST'),
    dbname=os.environ.get('POSTGRES_DB')
))


def truncate_sa_sources(scenario_sources):
    sa_sources = [
        source.subject
        for source in scenario_sources if isinstance(source.subject, pemi.SaDataSubject)
    ]
    for source in sa_sources:
        with source.engine.connect().begin() as trans:
            trans.connection.execute('TRUNCATE {}'.format(source.table))

#pylint: disable=unused-argument
@pytest.fixture(scope='module')
def db_schema_init():
    with sa_engine.connect() as conn:
        conn.execute(
            '''
            DROP TABLE IF EXISTS sales_fact;
            CREATE TABLE sales_fact (
              beer_id INT,
              name VARCHAR(80),
              sold_at DATE,
              quantity INT
            );
            '''
        )


@pytest.fixture
def db_case_clean(case, db_schema_init):
    if not case.scenario.has_run:
        truncate_sa_sources(case.scenario.sources.values())
        yield
        truncate_sa_sources(case.scenario.sources.values())
    else:
        yield
#pylint: enable=unused-argument


class BeersKeyFactory(factory.Factory):
    class Meta:
        model = dict
    id = factory.Sequence(lambda n: n)

sa_pipe = pemi.pipes.sa.SaSqlSourcePipe(
    engine=sa_engine,
    schema=pemi.Schema(
        beer_id=IntegerField(),
        name=StringField(),
        sold_at=DateField(),
        quantity=IntegerField()
    ),
    sql='SELECT * FROM sales_fact;'
)

sa_pipe.source(
    pemi.SaDataSubject,
    name='dummy_source',
    schema=sa_pipe.schema,
    engine=sa_engine,
    table='sales_fact'
)

with pt.Scenario(
    name='SaSqlSourcePipe',
    pipe=sa_pipe,
    factories={
        'beers': BeersKeyFactory
    },
    sources={
        'dummy_source': lambda pipe: pipe.sources['dummy_source']
    },
    targets={
        'main': lambda pipe: pipe.targets['main']
    },
    target_case_collectors={
        'main': pt.CaseCollector(subject_field='beer_id', factory='beers', factory_field='id')
    },
    usefixtures=['db_case_clean']
) as scenario:

    with scenario.case('it queries data stored') as case:
        ex_sales = pemi.data.Table(
            '''
            | beer_id | sold_at    | quantity |
            | -       | -          | -        |
            | {b[1]}  | 2017-01-01 | 3        |
            | {b[2]}  | 2017-01-02 | 2        |
            | {b[3]}  | 2017-01-03 | 5        |
            '''.format(
                b=scenario.factories['beers']['id']
            ),
            schema=scenario.sources['dummy_source'].schema
        )

        case.when(
            pt.when.example_for_source(scenario.sources['dummy_source'], ex_sales)
        ).then(
            pt.then.target_matches_example(scenario.targets['main'], ex_sales)
        )

    with scenario.case('it works with na') as case:
        ex_sales = pemi.data.Table(
            '''
            | beer_id | sold_at    | quantity |
            | -       | -          | -        |
            | {b[1]}  | 2017-01-01 | 3        |
            | {b[2]}  | 2017-01-02 |          |
            | {b[3]}  | 2017-01-03 | 5        |
            '''.format(
                b=scenario.factories['beers']['id']
            ),
            schema=scenario.sources['dummy_source'].schema
        )

        case.when(
            pt.when.example_for_source(scenario.sources['dummy_source'], ex_sales)
        ).then(
            pt.then.target_matches_example(scenario.targets['main'], ex_sales)
        )



from invoke import task

@task
def package(ctx):
    '''
    Package distribution to upload to PyPI
    '''
    ctx.run('rm -rf dist')
    ctx.run('python setup.py sdist')

@task
def package_deploy(ctx):
    '''
    Deploy package to PyPI
    '''
    ctx.run('twine upload dist/*')

@task
def requirements_compile(ctx):
    '''
    Compile Python requirements without upgrading.
    Docker images need to be rebuilt after running this (inv build).
    '''

    ctx.run('pip-compile requirements/base.in')
    ctx.run('pip-compile requirements/doc.in')
    ctx.run('pip-compile requirements/all.in')

@task
def requirements_upgrade(ctx):
    '''
    Compile Python requirements with upgrading.
    Docker images need to be rebuilt after running this (inv build).
    '''

    ctx.run('pip-compile -U requirements/base.in')
    ctx.run('pip-compile -U requirements/doc.in')
    ctx.run('pip-compile -U requirements/all.in')

@task
def build(ctx):
    'Build docker images'
    ctx.run('docker-compose build')

@task
def up(ctx):
    'Start up development environment'
    ctx.run('docker-compose up -d')

@task
def down(ctx):
    'Shut down development environment'
    ctx.run('docker-compose down')

@task(down, up)
def restart(ctx):
    'Restart development environment'

@task
def logs(ctx):
    'Follow docker logs'
    ctx.run('docker-compose logs -f')

DEFAULT_TEST_OPTS = '-s -x -vv --tb=short --color=yes'
@task(help={
    'tests': 'Specify the test you want to run - e.g., tests/test_fields.py',
    'opts': 'Set the options passed to pytest (default: {})'.format(DEFAULT_TEST_OPTS),
    'ignore': 'Specify the tests you want to ignore - e.g., tests/test_spark_job.py',
})
def test(ctx, opts=DEFAULT_TEST_OPTS, tests='tests', ignore=None):
    'Runs the test suite.  User can specifiy pytest options to run specific tests.'

    if ignore:
        opts += ' --ignore={}'.format(ignore)
    ctx.run('docker-compose run app pytest {} {}'.format(opts, tests))


@task(help={
    'all': 'Run all linters (default)',
    'app': 'Just run the app linter (pemi)',
    'tests': 'Just run the tests linter',
    'files': 'Specify files -- eg., pemi/fields.py'
})
def lint(ctx, all=False, app=False, tests=False, files=None):
    'Checks code quality with pylint'
    all = all or not (app or tests)

    if all or app:
        opts = files or 'pemi'
        ctx.run('docker-compose run app pylint --rcfile pemi/.pylintrc {}'.format(opts))

    if all or tests:
        opts = files or 'tests'
        ctx.run('docker-compose run app pylint --rcfile tests/.pylintrc {}'.format(opts))


@task
def doc_generate(ctx):
    'Generates documents locally and writes to docs/_build/html'
    ctx.run('docker-compose run -w /app/docs --rm app sphinx-build -E . _build/html')

@task
def ps(ctx):
    'View environment containers'
    ctx.run('docker-compose ps')

@task
def shell(ctx):
    'Start a shell running in the app container'
    ctx.run('docker-compose run --rm app /bin/bash')



import decimal
import datetime
import json

from functools import wraps

import dateutil

import pemi.transforms

__all__ = [
    'StringField',
    'IntegerField',
    'FloatField',
    'DateField',
    'DateTimeField',
    'BooleanField',
    'DecimalField',
    'JsonField'
]


class CoercionError(ValueError): pass
class DecimalCoercionError(ValueError): pass

def convert_exception(fun):
    @wraps(fun)
    def wrapper(self, value):
        try:
            coerced = fun(self, value)
        except Exception as err:
            raise CoercionError('Unable to coerce value "{}" to {}: {}: {}'.format(
                value,
                self.__class__.__name__,
                err.__class__.__name__,
                err
            ))
        return coerced
    return wrapper

#pylint: disable=too-few-public-methods
class Field:
    '''
    A field is a thing that is inherited
    '''

    def __init__(self, name=None, **metadata):
        self.name = name
        self.metadata = metadata

        default_metadata = {'null': None}
        self.metadata = {**default_metadata, **metadata}
        self.null = self.metadata['null']

    @convert_exception
    def coerce(self, value):
        raise NotImplementedError

    def __str__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.__dict__.__str__())

    def __eq__(self, other):
        return type(self) is type(other) \
            and self.metadata == other.metadata \
            and self.name == other.name


class StringField(Field):
    def __init__(self, name=None, **metadata):
        metadata['null'] = metadata.get('null', '')
        super().__init__(name=name, **metadata)

    @convert_exception
    def coerce(self, value):
        if pemi.transforms.isblank(value):
            return self.null
        return str(value).strip()


class IntegerField(Field):
    def __init__(self, name=None, **metadata):
        super().__init__(name=name, **metadata)
        self.coerce_float = self.metadata.get('coerce_float', False)

    @convert_exception
    def coerce(self, value):
        if pemi.transforms.isblank(value):
            return self.null
        if self.coerce_float:
            return int(float(value))
        return int(value)


class FloatField(Field):
    @convert_exception
    def coerce(self, value):
        if pemi.transforms.isblank(value):
            return self.null
        return float(value)


class DateField(Field):
    def __init__(self, name=None, **metadata):
        super().__init__(name=name, **metadata)
        self.format = self.metadata.get('format', '%Y-%m-%d')
        self.infer_format = self.metadata.get('infer_format', False)


    @convert_exception
    def coerce(self, value):
        if hasattr(value, 'strip'):
            value = value.strip()

        if pemi.transforms.isblank(value):
            return self.null
        return self.parse(value)

    def parse(self, value):
        if isinstance(value, datetime.datetime):
            return value.date()
        if isinstance(value, datetime.date):
            return value
        if not self.infer_format:
            return datetime.datetime.strptime(value, self.format).date()
        return dateutil.parser.parse(value).date()

class DateTimeField(Field):
    def __init__(self, name=None, **metadata):
        super().__init__(name=name, **metadata)
        self.format = self.metadata.get('format', '%Y-%m-%d %H:%M:%S')
        self.infer_format = self.metadata.get('infer_format', False)


    @convert_exception
    def coerce(self, value):
        if hasattr(value, 'strip'):
            value = value.strip()

        if pemi.transforms.isblank(value):
            return self.null
        return self.parse(value)

    def parse(self, value):
        if isinstance(value, datetime.datetime):
            return value
        if isinstance(value, datetime.date):
            return datetime.datetime.combine(value, datetime.time.min)
        if not self.infer_format:
            return datetime.datetime.strptime(value, self.format)
        return dateutil.parser.parse(value)


class BooleanField(Field):
    # when defined, the value of unknown_truthiness is used when no matching is found
    def __init__(self, name=None, **metadata):
        super().__init__(name=name, **metadata)

        self.true_values = self.metadata.get(
            'true_values',
            ['t', 'true', 'y', 'yes', 'on', '1']
        )
        self.false_values = self.metadata.get(
            'false_values',
            ['f', 'false', 'n', 'no', 'off', '0']
        )

    @convert_exception
    def coerce(self, value):
        if hasattr(value, 'strip'):
            value = value.strip()

        if isinstance(value, bool):
            return value
        if pemi.transforms.isblank(value):
            return self.null
        return self.parse(value)

    def parse(self, value):
        value = str(value).lower()
        if value in self.true_values:
            return True
        if value in self.false_values:
            return False
        if 'unknown_truthiness' in self.metadata:
            return self.metadata['unknown_truthiness']
        raise ValueError('Not a boolean value')

class DecimalField(Field):
    def __init__(self, name=None, **metadata):
        super().__init__(name=name, **metadata)
        self.precision = self.metadata.get('precision', 16)
        self.scale = self.metadata.get('scale', 2)
        self.truncate_decimal = self.metadata.get('truncate_decimal', False)
        self.enforce_decimal = self.metadata.get('enforce_decimal', True)

    @convert_exception
    def coerce(self, value):
        if pemi.transforms.isblank(value):
            return self.null
        return self.parse(value)

    def parse(self, value):
        dec = decimal.Decimal(str(value))

        if dec != dec: #pylint: disable=comparison-with-itself
            return dec

        if self.truncate_decimal:
            dec = round(dec, self.scale)

        if self.enforce_decimal:
            detected_precision = len(dec.as_tuple().digits)
            detected_scale = -dec.as_tuple().exponent

            if detected_precision > self.precision:
                msg = ('Decimal coercion error for "{}".  ' \
                    + 'Expected precision: {}, Actual precision: {}').format(
                        dec, self.precision, detected_precision
                    )
                raise DecimalCoercionError(msg)
            if detected_scale > self.scale:
                msg = ('Decimal coercion error for "{}".  ' \
                    + 'Expected scale: {}, Actual scale: {}').format(
                        dec, self.scale, detected_scale
                    )
                raise DecimalCoercionError(msg)

        return dec

class JsonField(Field):
    @convert_exception
    def coerce(self, value):
        if pemi.transforms.isblank(value):
            return self.null

        try:
            return json.loads(value)
        except TypeError:
            return value

#pylint: enable=too-few-public-methods



import warnings
import re

def pytest_generate_tests(metafunc):
    if not hasattr(metafunc.function, 'pytestmark'):
        return

    pymarks = {mark.name: mark for mark in metafunc.function.pytestmark}
    if 'scenario' in pymarks:
        args = pymarks['scenario'].args
        scenario = args[0]

        if args[1] is not None:
            warnings.warn(
                'WARNING - Only a subset of cases are selected: {}'.format(args)
            )
            re_tag = re.compile(args[1])
            for name, _ in list(scenario.cases.items()):
                if not re_tag.search(name):
                    del scenario.cases[name]

        metafunc.parametrize(
            ['case'],
            [[case] for case in scenario.cases.values()],
            ids=[case.name for case in scenario.cases.values()]
        )



'''

Testing is described with examples in :doc:`tests`.

.. autoclass:: Scenario
  :members:

.. autoclass:: Case
  :members:

.. autoclass:: when
  :members:

.. autoclass:: then
  :members:

'''

from collections import namedtuple
from collections import OrderedDict

import os
import sys
import inspect

import pytest
import pandas as pd

import pemi
import pemi.data

from pemi.tabular import PemiTabular

# Imported here for backwards compatiblity
from pemi.pipe import mock_pipe #pylint: disable=unused-import


#pylint: disable=too-many-lines


pd.set_option('display.expand_frame_repr', False)

class KeyFactoryFieldError(Exception): pass
class CaseStructureError(Exception): pass
class NoTargetCaseCollectorError(Exception): pass
class UnableToFindCaseError(Exception): pass
class NoTargetDataError(AssertionError): pass

def assert_frame_equal(actual, expected, **kwargs):
    try:
        pd.util.testing.assert_frame_equal(actual, expected, **kwargs)
    except AssertionError as err:
        msg = str(err)
        msg += '\nActual:\n{}'.format(actual)
        msg += '\nExpected:\n{}'.format(expected)
        raise AssertionError(msg)

def assert_series_equal(actual, expected, **kwargs):
    actual.reset_index(drop=True, inplace=True)
    expected.reset_index(drop=True, inplace=True)
    try:
        pd.util.testing.assert_series_equal(actual, expected, **kwargs)
    except AssertionError as err:
        msg = str(err)
        msg += '\nActual:\n{}'.format(actual)
        msg += '\nExpected:\n{}'.format(expected)
        raise AssertionError(msg)

class when: #pylint: disable=invalid-name
    #pylint: enable=invalid-name
    '''
    Contains methods used to set up conditions for a testing case.
    '''

    @staticmethod
    def source_field_has_value(source, field, value):
        '''
        Sets the value of a specific field to a specific value.

        Args:
            source (scenario.sources[]): The scenario source data subject.

            field (str): Name of field.

            value (str, iter): Value to set for the field.

        Examples:
            Set the value of the field ``name`` to the string value ``Buffy`` in
            the scenario source ``main``::

                case.when(
                    when.source_field_has_value(scenario.sources['main'], 'name', 'Buffy')
                )
        '''

        def _when(case):
            nrecords = len(source[case].data)
            if hasattr(value, '__next__'):
                values = pd.Series([next(value) for i in range(nrecords)])
            else:
                values = pd.Series([value]*nrecords)

            if field in source.schema:
                values = values.apply(
                    lambda v: v if hasattr(v, '__pemi_test_no_coerce__') \
                        else source.schema[field].coerce(v)
                )

            source[case].data[field] = values

        return _when

    @staticmethod
    def source_fields_have_values(source, mapping):
        '''
        Sets the value of a multiples fields to a specific values.

        Args:
            source (scenario.sources[]): The scenario source data subject.

            mapping (dict): Dictionary where the keys are the names of fields and the values
                are the values those fields are to be set to.

        Examples:
            Set the value of the field ``name`` to the string value ``Buffy`` and
            the value of the field ``vampires_slain`` to ``133`` in
            the scenario source ``main``::

                case.when(
                    when.source_fields_have_values(
                        scenario.sources['main'],
                        {
                            'name': 'Buffy',
                            'vampires_slain': 133
                        }
                    )
                )
        '''

        def _when(case):
            for field, value in mapping.items():
                nrecords = len(source[case].data)
                if hasattr(value, '__next__'):
                    values = pd.Series([next(value) for i in range(nrecords)])
                else:
                    values = pd.Series([value]*nrecords)

                if field in source.schema:
                    values = values.apply(
                        lambda v, f=field: v if hasattr(v, '__pemi_test_no_coerce__') \
                            else source.schema[f].coerce(v)
                    )

                source[case].data[field] = values


        return _when

    @staticmethod
    def example_for_source(source, table):
        """
        Set specific rows and columns to specific values.

        Args:
            source (scenario.sources[]): The scenario source data subject.

            table (pemi.data.Table): Pemi data table to use for specifying data.

        Example:
            Given a Pemi data table, specify rows and columns for the source `main`::

                case.when(
                    when.example_for_source(
                        scenario.sources['main'],
                        pemi.data.Table(
                            '''
                            | id       | name  |
                            | -        | -     |
                            | {sid[1]} | Spike |
                            | {sid[2]} | Angel |
                            '''.format(
                                sid=scenario.factories['vampires']['id']
                            )
                        )
                    )
                )

        """

        def _when(case):
            source[case].data = table.df

        return _when

    @staticmethod
    def source_conforms_to_schema(source, key_factories=None):
        '''
        Creates 3 records and fills out data for a data source subject
        that conforms to the data types specified by the data
        subject's schema.

        Args:
            source (scenario.sources[]): The scenario source data subject.

            key_factories (dict): A dictionary where the keys are the names of fields
            and the values are the field value generator originating from a scenario
            key factory (just see the example ;)).

        Example:

            For the source subject 'main', this will generate faked data that conforms
            to the schema defined for main.  It will also populate the ``id`` field
            with values generated from the ``id`` field in the ``vampires`` factory::

                case.when(
                    when.source_conforms_to_schema(
                        scenario.sources['main'],
                        {'id': scenario.factories['vampires']['id']}
                    )
                )

        '''


        key_factories = key_factories or {}
        ntestrows = 3

        key_values = {
            field: [factory[os.urandom(32)] for _ in range(ntestrows)]
            for field, factory in key_factories.items()
        }

        def _when(case):
            data = pemi.data.Table(
                nrows=ntestrows,
                schema=source.subject.schema
            ).df

            for field, values in key_values.items():
                data[field] = pd.Series(values, index=data.index)

            source[case].data = data
        return _when


class then: #pylint: disable=invalid-name
    #pylint: enable=invalid-name

    '''
    Contains methods used to test that actual outcome is equal to expected outcome.
    '''

    @staticmethod
    def target_field_has_value(target, field, value):
        '''
        Asserts that a specific field has a specific value.

        Args:
            target (scenario.targets[]): The scenario target data subject.

            field (str): Name of field.

            value (str): Value of the field that is expected.

        Examples:
            Asserts that the value of the field ``name`` is set to the string value ``Buffy`` in
            the scenario target ``main``::

                case.then(
                    then.target_field_has_value(scenario.targets['main'], 'name', 'Buffy')
                )
        '''

        def _then(case):
            if len(target[case].data[field]) < 1:
                raise NoTargetDataError('Target has no data for case')

            target_data = pd.Series(list(target[case].data[field]))
            expected_data = pd.Series([value] * len(target_data), index=target_data.index)

            assert_series_equal(target_data, expected_data,
                                check_names=False,
                                check_dtype=False,
                                check_datetimelike_compat=True)

        return _then

    @staticmethod
    def target_fields_have_values(target, mapping):
        '''
        Asserts that multiple fields have specific values.

        Args:
            target (scenario.targets[]): The scenario target data subject.

            mapping (dict): Dictionary where the keys are the names of fields and the values
                are the expected values those fields.

        Examples:
            Asserts that the value of the field ``name`` is the string value ``Buffy`` and
            the value of the field ``vampires_slain`` is ``133`` in
            the scenario target ``main``::

                case.then(
                    then.target_fields_have_values(
                        scenario.targets['main'],
                        {
                            'name': 'Buffy',
                            'vampires_slain': 133
                        }
                    )
                )
        '''

        def _then(case):
            if len(target[case].data) < 1:
                raise NoTargetDataError('Target has no data for case')

            actual = target[case].data[list(mapping.keys())]
            expected = pd.DataFrame(index=actual.index)
            for k, v in mapping.items():
                expected[k] = pd.Series([v] * len(actual), index=actual.index)

            assert_frame_equal(actual, expected, check_names=False, check_dtype=False)
        return _then

    @staticmethod
    def target_matches_example(target, expected_table, by=None):
        """
        Asserts that a given target matches an example data table

        Args:
            target (scenario.targets[]): The scenario target data subject.

            expected_table (pemi.data.Table): Expected result data.  If the table
                has fewer columns than the pipe generates, those extra columns are
                not considered in the comparison.

            by (list): A list of field names to sort the result data by before
                performing the comparison.

        Examples:
            Asserts that the scenario target ``main`` conforms to the expected data::

                case.then(
                    then.target_matches_example(
                        scenario.targets['main'],
                        pemi.data.Table(
                            '''
                            | id       | name  |
                            | -        | -     |
                            | {sid[1]} | Spike |
                            | {sid[2]} | Angel |
                            '''.format(
                                sid=scenario.factories['vampires']['id']
                            )
                        ),
                        by=['id'] #esp important if the ids are generated randomly
                    )
                )
        """


        subject_fields = expected_table.defined_fields

        def _then(case):
            expected = expected_table.df[subject_fields]
            actual = target[case].data[subject_fields]

            if by:
                expected = expected.sort_values(by).reset_index(drop=True)
                actual = actual.sort_values(by).reset_index(drop=True)
            else:
                expected = expected.reset_index(drop=True)
                actual = actual.reset_index(drop=True)

            assert_frame_equal(actual, expected, check_names=False, check_dtype=False)

        return _then

    @staticmethod
    def field_is_copied(source, source_field, target, target_field, by=None, #pylint: disable=too-many-arguments
                        source_by=None, target_by=None):
        '''
        Asserts that a field value is copied from the source to the target.

        Args:
            source (scenario.sources[]): The scenario source data subject.

            source_field (str): The name of the source field.

            target (scenario.targets[]): The scenario target data subject.

            target_field (str): The name of the target field.

            by (list): A list of field names to sort the data by before
                performing the comparison.

            source_by (list): A list of field names to sort the source data by before
                performing the comparison (uses ``by`` if not given).

            target_by (list): A list of field names to sort the target data by before
                performing the comparison (uses ``by`` if not given).

        Examples:
            Asserts that the value of the source field ``name`` is copied to the
            target field ``slayer_name``::

                case.then(
                    then.field_is_copied(
                        scenario.sources['main'], 'name',
                        scenario.targets['main'], 'slayer_name',
                        by=['id']
                    )
                )
        '''

        source_by = source_by or by
        target_by = target_by or by or source_by

        def _then(case):
            if source_by:
                expected = source[case].data.sort_values(source_by)\
                                            .reset_index(drop=True)[[source_field]]
                actual = target[case].data.sort_values(target_by)\
                                          .reset_index(drop=True)[[target_field]]
            else:
                expected = source[case].data[[source_field]]
                actual = target[case].data[[target_field]]

            try:
                assert_series_equal(actual[target_field], expected[source_field], check_names=False)
            except AssertionError as err:
                raise AssertionError(
                    'Source field {} not copied to target field {}: {}'.format(
                        source_field, target_field, err
                    )
                )

        return _then

    @staticmethod
    def fields_are_copied(source, target, mapping, by=None, source_by=None, target_by=None): #pylint: disable=too-many-arguments
        '''
        Asserts that various field values are copied from the source to the target.

        Args:
            source (scenario.sources[]): The scenario source data subject.

            target (scenario.targets[]): The scenario target data subject.

            mapping (list): A list of tuples.  Each tuple contains the source field name
                and target field name, in that order.

            by (list): A list of field names to sort the data by before
                performing the comparison.

            source_by (list): A list of field names to sort the source data by before
                performing the comparison (uses ``by`` if not given).

            target_by (list): A list of field names to sort the target data by before
                performing the comparison (uses ``by`` if not given).

        Examples:
            Asserts that the value of the source field ``name`` is copied to the
            target field ``slayer_name`` and ``num`` is copied to ``vampires_slain``::

                case.then(
                    then.fields_are_copied(
                        scenario.sources['main'],
                        scenario.targets['main'],
                        [
                            ('name', 'slayer_name'),
                            ('num', 'vampires_slain')
                        ],
                        by=['id']
                    )
                )
        '''


        source_fields = list({m[0] for m in mapping})
        target_fields = list({m[1] for m in mapping})

        source_by = source_by or by
        target_by = target_by or by or source_by

        def _then(case):
            if source_by:
                expected = source[case].data.sort_values(source_by)\
                                            .reset_index(drop=True)[source_fields]
                actual = target[case].data.sort_values(target_by)\
                                          .reset_index(drop=True)[target_fields]
            else:
                expected = source[case].data[source_fields]
                actual = target[case].data[target_fields]

            for source_field, target_field in mapping:
                try:
                    assert_series_equal(actual[target_field], expected[source_field],
                                        check_names=False, check_dtype=False)
                except AssertionError as err:
                    raise AssertionError(
                        'Source field {} not copied to target field {}: {}'.format(
                            source_field, target_field, err
                        )
                    )

        return _then


    @staticmethod
    def target_does_not_have_fields(target, fields):
        '''
        Asserts that the target does not have certain fields.

        Args:
            target (scenario.targets[]): The scenario target data subject.

            fields (list): List of field names that should not be on the target.

        Examples:
            Asserts that the scenario target ``main`` does not have the fields
            ``sparkle_factor`` or ``is_werewolf``::

                case.then(
                    then.target_does_not_have_fields(
                        scenario.targets['main'],
                        ['sparkle_factor', 'is_werewolf']
                    )
                )
        '''


        def _then(case):
            unexpected_fields = set(fields) & set(target[case].data.columns)
            if len(unexpected_fields) > 0:
                raise AssertionError(
                    "The fields '{}' were not expected to be found in the target".format(
                        unexpected_fields
                    )
                )

        return _then

    @staticmethod
    def target_has_fields(target, fields, only=False):
        '''
        Asserts that the target has certain fields.

        Args:
            target (scenario.targets[]): The scenario target data subject.

            fields (list): List of field names that should not be on the target.

            only (bool): Specifies whether the target should only have the fields listed.  Raises
                an exception if there are additional fields.

        Examples:
            Asserts that the scenario target ``main`` only has the fields
            ``name`` and ``vampires_slain``::

                case.then(
                    then.target_has_fields(
                        scenario.targets['main'],
                        ['name', 'vampires_slain'],
                        only=True
                    )
                )
        '''

        def _then(case):
            missing_fields = set(fields) - set(target[case].data.columns)
            extra_fields = set(target[case].data.columns) - set(fields)

            if len(missing_fields) > 0:
                raise AssertionError(
                    "The fields '{}' were expected to be found in the target".format(
                        missing_fields
                    )
                )

            if len(extra_fields) > 0 and only:
                raise AssertionError(
                    "The fields '{}' were not expected to be found on the target".format(
                        extra_fields
                    )
                )

        return _then

    @staticmethod
    def target_is_empty(target):
        '''
        Asserts that the target has no records.

        Args:
            target (scenario.targets[]): The scenario target data subject.

        Examples:
            Asserts that the scenario target ``errors`` does not have any records::

                case.then(then.target_is_empty(scenario.targets['errors'])
        '''

        def _then(case):
            nrecords = len(target[case].data)
            if nrecords != 0:
                raise AssertionError(
                    'Expecting target to be empty, found {} records'.format(
                        nrecords
                    )
                )

        return _then

    @staticmethod
    def target_has_n_records(target, expected_n):
        '''
        Asserts that the target has a specific number of records.

        Args:
            target (scenario.targets[]): The scenario target data subject.

            expected_n (int): The number of records expected.

        Examples:
            Asserts that the scenario target ``main`` has 3 records::

                case.then(then.target_has_n_records(scenario.targets['main'], 3)
        '''

        def _then(case):
            nrecords = len(target[case].data)
            if nrecords != expected_n:
                raise AssertionError(
                    'Excpecting target to have {} records, found {} records'.format(
                        expected_n, nrecords
                    )
                )

        return _then


class SubscriptableLambda: #pylint: disable=too-few-public-methods
    '''
    Used to help with putting specific values in example data tables.

    Args:
        func (func): Some python function you want to access as subscriptable.

    Examples:
        In the simplest form::

            sl = SubscriptableLambda(lambda v: v + 10)
            sl[3] #=> 13


        This class is useful in tests when creating complex methods that need to be used
        int table data::

            payload = pt.SubscriptableLambda(lambda ref: json.dumps({
                'external_id': scenario.factories['students']['external_id'][ref]
            }))

            response = pt.SubscriptableLambda(lambda ref: json.dumps([{
                'itk-api': [
                    {'resource_uuid': scenario.factories['students']['uuid'][ref]}
                ]
            }]))

            ex_create_response = pemi.data.Table(
                """
                | payload             | response             |
                | -                   | -                    |
                | {payload[created1]} | {response[created1]} |
                | {payload[created2]} | {response[created2]} |
                | {payload[created3]} | {response[created3]} |
                | {payload[created4]} | {response[created4]} |
                """.format(
                    payload=payload,
                    response=response
                ),
                schema=pemi.Schema(
                    payload=JsonField(),
                    response=JsonField()
                )
            )
    '''

    def __init__(self, func):
        self.func = func

    def __getitem__(self, key=None):
        return self.func(key)


CaseCollector = namedtuple('CaseCollector', ['subject_field', 'factory', 'factory_field'])

class DuplicateScenarioError(Exception): pass
class DuplicateCaseError(Exception): pass

class Scenario: #pylint: disable=too-many-instance-attributes, too-many-arguments
    '''
    A **Scenario** describes the transformation that is being tested
    (a Pemi pipe), and the data sources and targets that are the
    subject of the test.  Scenarios are composed of one more **Cases**.

    Args:
        name (str): The name of a scenario.  Multiple scenarios may be present in a file,
            but the names of each scenario must be unique.

        pipe (pemi.Pipe): The Pemi pipe that is the main subject of the test.  Test
            data will be provided to the sources of the pipe (defined below), and the pipe
            will be executed.  Note that the pipe is only executed once per scenario.

        flow (str): The name of the method used to execute the pipe (default: `flow`).

        factories(dict): A dictionary where the keys are the names of factories and
            the values are FactoryBoy factories that will be used to generate unique keys.

        sources (dict): A dictionary where the keys are the names of sources that will
            be the subjects of testing.  The values are methods that accept the pipe
            referenced in the **pipe** argument above and return the data subject that
            will be used as a source.

        targets (dict): A dictionary where the keys are the names of targets that will
            be the subjects of testing.  The values are methods that accept the pipe
            referenced in the **pipe** argument above and return the data subject that
            will be used as a target.

        target_case_collectors (dict): A dictionary where the keys are the names of the
            targets that will be the subjects of testing.  The values are ``CaseCollector``
            objects that tie a field in the scenario's target to the field in a given factory.
            Every named target needs to have a case collector.

        selector (str): A string representing a regular expression.  Any case names that
            **do not** match this regex will be excluded from testing.

        usefixtures (str): Name of a Pytest fixture to use for the scenario.  Often used
            for database setup/teardown options.

    '''

    def __init__(self, name, pipe, factories, sources, targets, target_case_collectors,
                 flow='flow', selector=None, usefixtures=None):
        self.name = name
        self.pipe = pipe
        self.flow = flow
        self.factories = self._setup_factories(factories)
        self.sources = self._setup_subjects(sources)
        self.targets = self._setup_subjects(targets)
        self.target_case_collectors = target_case_collectors
        self.selector = selector
        self.usefixtures = usefixtures or []

        self.cases = OrderedDict()
        self.has_run = False

    def _register_test(self, module_name):
        @pytest.mark.usefixtures(*self.usefixtures)
        @pytest.mark.scenario(self, self.selector)
        def test_scenario(case):
            case.assert_case()
        test_attr = 'testScenario:{}'.format(self.name)
        if hasattr(sys.modules[module_name], test_attr):
            raise DuplicateScenarioError(
                'Scenario names must be unique to a module.  '
                'Duplicate detected: {}'.format(test_attr)
            )
        setattr(sys.modules[module_name], test_attr, test_scenario)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        current_frame = inspect.currentframe()
        calling_module = inspect.getouterframes(current_frame)[1].frame.f_locals['__name__']
        self._register_test(calling_module)

    @staticmethod
    def _setup_factories(factories):
        return {name: KeyFactory(factory) for name, factory in factories.items()}

    def _setup_subjects(self, subjects):
        return {name: TestSubject(subject(self.pipe), name) for name, subject in subjects.items()}

    def case(self, name):
        if name in self.cases:
            raise DuplicateCaseError(
                'Case names must be unique to a scenario.  '
                'Duplicate case detected in scenario "{}": "{}"'.format(
                    self.cases[name].scenario.name, name
                )
            )

        case = Case(name, self)

        for factory in self.factories.values():
            factory.next_case(case)

        self.cases[name] = case
        return case

    def run(self):
        if self.has_run:
            return

        self.setup_cases()
        getattr(self.pipe, self.flow)()
        self.collect_results()

        self.has_run = True

    def setup_cases(self):
        for case in self.cases.values():
            case.setup()

        self.load_test_data()

    def load_test_data(self):
        for _, source in self.sources.items():
            if len(source.data.values()) > 0:
                all_case_data = pd.concat(
                    [cd.data for cd in source.data.values()],
                    ignore_index=True,
                    sort=False
                )
                source.subject.from_pd(all_case_data)

    def collect_results(self):
        for target_name, target in self.targets.items():
            all_target_data = target.subject.to_pd()
            for case in self.cases.values():
                target[case].data = pd.DataFrame(columns=all_target_data.columns)

            try:
                collector = self.target_case_collectors[target_name]
            except KeyError:
                raise NoTargetCaseCollectorError(
                    'No case collector defined for target {}'.format(target_name)
                )

            if len(all_target_data) > 0:
                all_target_data['__pemi_case__'] = all_target_data[collector.subject_field].apply(
                    self.factories[collector.factory].case_lookup(collector)
                )
                for case, df in all_target_data.groupby(['__pemi_case__'], sort=False):
                    del df['__pemi_case__']
                    target[case].data = df

class Case:
    '''
    A **Case** is a set of **Conditions** and **Expectations** that describe
    how the pipe is supposed to function.

    Args:

        name (str): The name of the case.  The names of cases within a scenario must be unique.

        scenario (pemi.testing.Scenario): The scenario object that this case is associated with.
    '''

    def __init__(self, name, scenario):
        self.name = name
        self.scenario = scenario
        self.whens = []
        self.thens = []
        self.expected_exception = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def when(self, *funcs):
        '''
        Accepts a list of functions that are used to set up the data for a specific case.
        Each of the functions should accept one argument, which is the case object.
        See pemi.testing.when for examples.
        '''

        self.whens.extend(funcs)
        return self

    def then(self, *funcs):
        '''
        Accepts a list of functions that are used to test the result data for a specific case.
        Each of the functions should accept one argument, which is the case object.
        See pemi.testing.then for examples.
        '''

        self.thens.extend(funcs)
        return self

    def expect_exception(self, exception):
        '''
        Used to indicate that the test case is expected to fail with exception ``exception``.
        If the test case raises this exception, then it will pass.  If it does not raise the
        exception, then it will fail.
        '''

        self.expected_exception = exception

    def setup(self):
        for i_when in self.whens:
            i_when(self)

    def _assert_then(self, i_then):
        if self.expected_exception:
            with pytest.raises(self.expected_exception):
                i_then(self)
        else:
            i_then(self)

    def assert_case(self):
        self.scenario.run()

        try:
            if len(self.thens) < 1:
                raise CaseStructureError
            for i_then in self.thens:
                self._assert_then(i_then)

        except AssertionError:
            errors_tbl = PemiTabular()
            msg = '\nAssertion Error for {}'.format(self)
            for name, source in self.scenario.sources.items():
                source_df = source.subject.to_pd()
                msg += '\nSource {}:\n{}'.format(name, source_df)
                errors_tbl.add(df=source_df, df_name="\nSource {}".format(name))
            for name, target in self.scenario.targets.items():
                target_df = target.subject.to_pd()
                msg += '\nTarget {}:\n{}'.format(name, target_df)
                errors_tbl.add(df=target_df, df_name="\nTarget {}".format(name))
            errors_tbl.render(file='pemi-errors.html')
            raise AssertionError(msg)
        except CaseStructureError:
            msg = '\nCase Structure Error for {}'.format(self)
            msg += '\tNo .then clause found in test case'
            raise CaseStructureError(msg)

    def __str__(self):
        return "<Case '{}' ({})>".format(self.name, id(self))






class KeyFactoryField: #pylint: disable=too-few-public-methods
    '''
    For internal use only.

    Used to access a particular field from a given key factory.

    Example::

        kff = KeyFactoryField(keyfactory, 'id')
        kff[2] #=> Returns the 'id' field from the keyfactory instance referenced by then integer 2
    '''

    def __init__(self, keyfactory, field):
        self.keyfactory = keyfactory
        self.field = field

    def __getitem__(self, ref=None):
        '''
        Args:
          ref (Object) - Any hashable object used to reference a particular key factory instance.
        '''
        instance = self.keyfactory.instance(ref)
        if self.field in instance:
            return instance[self.field]

        raise KeyFactoryFieldError(
            'Key field "{}" not defined for factory {}'.format(
                self.field, self.keyfactory.factory
            )
        )

class KeyFactory:
    '''
    For internal use only.

    Wrapper around a FactoryBoy factory that caches any factory instances created.

    Example::

        class BeersKeyFactory(factory.Factory):
            class Meta:
                model = dict
            beer_id = factory.Sequence(lambda n: n)

        keyfactory = KeyFactory(BeersKeyFactory)

        keyfactory.instance('a') #=> generates a new instance of BeersKeyFactory cached with key 'a'
        keyfactory['beer_id']['a'] #=> Returns field 'beer_id' from cache referenced by 'a'

        keyfactory.instance('z') #=> generates a new instance of BeersKeyFactory cached with key 'z'
        keyfactory['beer_id']['z'] #=> Returns field 'beer_id' from cache referenced by 'z'
    '''

    def __init__(self, factory):
        self.factory = factory
        self.case = None
        self.cached = {}
        self.next_case(None)

    def next_case(self, case):
        self.case = case
        self.cached[self.case] = {}

    def __getitem__(self, field):
        return KeyFactoryField(self, field)

    def instance(self, ref=None):
        ref = os.urandom(32) if ref is None else ref

        if ref not in self.cached[self.case]:
            self.cached[self.case][ref] = self.factory()
        return self.cached[self.case][ref]

    def case_lookup(self, collector):
        lkp = {}
        for case_id, cached_keys in self.cached.items():
            for keys in cached_keys.values():
                if collector.factory_field not in keys:
                    raise KeyFactoryFieldError(
                        '"{}" is not a known factory fields for {}'.format(
                            collector.factory_field, self.factory
                        )
                    )
                lkp[keys[collector.factory_field]] = case_id

        def _lkp(v):
            try:
                return lkp[v]
            except KeyError as _err:
                raise UnableToFindCaseError(
                    'Unable to associate field "{}" and value "{}" with a case'.format(
                        collector.subject_field, v
                    )
                )

        return _lkp



class CaseData: #pylint: disable=attribute-defined-outside-init,too-few-public-methods
    '''
    For internal use only.

    Creates a dataframe for a specific test case.
    All of the ``when`` conditions in a case will add or modify this dataframe
    before it gets concatenated with other cases and run through the pipe that
    is the subject of the scenario.
    '''

    def __init__(self, case, test_subject):
        self.case = case
        self.test_subject = test_subject

    @property
    def data(self):
        if not hasattr(self, '_data'):
            cols = self.test_subject.subject.schema.keys()
            self._data = pd.DataFrame([[None] * len(cols)], columns=cols)

        return self._data

    @data.setter
    def data(self, value):
        self._data = value

class TestSubject: #pylint: disable=too-few-public-methods
    '''
    For internal use only.

    Wrapper around an actual pipe data subject.  Used to set or fetch case-specific data records.
    '''

    def __init__(self, subject, name):
        self.subject = subject
        self.name = name
        self.data = {}
        self.schema = self.subject.schema

    def __getitem__(self, case):
        if case not in self.data:
            self.data[case] = CaseData(case, self)
        return self.data[case]
